import { PatchManager } from './utils/patchManager.js';
import NavigationBarFeature from './features/navigationBar/navigationBarFeature.js';
import OskKeyPopupsFeature from './features/osk/oskKeyPopupsFeature.js';
import Clutter from 'gi://Clutter';
import { NotificationGesturesFeature } from './features/notifications/notificationGesturesFeature.js';
import { Extension } from 'resource:///org/gnome/shell/extensions/extension.js';
import { initSettings, uninitSettings } from './features/preferences/backend.js';
import { FloatingScreenRotateButtonFeature } from './features/screenRotateUtils/floatingScreenRotateButtonFeature.js';
import { Delay } from './utils/delay.js';
import { assetsGResourceFile } from './config.js';
import { settings } from './settings.js';
import Gio from 'gi://Gio';
import DonationsFeature from './features/donations/donationsFeature.js';

class TouchUpExtension extends Extension {
    static instance;
    pm;
    developmentTools;
    navigationBar;
    oskKeyPopups;
    floatingScreenRotateButton;
    virtualTouchpad;
    notificationGestures;
    donations;
    enable() {
        // This is the root patch manager of which all other patch managers are descendents:
        this.pm = new PatchManager("root");
        // Load assets:
        this.pm.patch(() => {
            const assets = Gio.resource_load(this.dir.get_child(assetsGResourceFile).get_path());
            Gio.resources_register(assets);
            return () => Gio.resources_unregister(assets);
        }, 'load-and-register-assets');
        // Initialize settings:
        this.pm.patch(() => {
            initSettings(this.getSettings());
            return () => uninitSettings();
        }, 'init-settings');
        TouchUpExtension.instance = this;
        // This is the entry point for all features of this extension:
        this.defineFeatures();
        // Sync ui on touch-mode and monitor changes:
        this.pm.connectTo(Clutter.get_default_backend().get_default_seat(), 'notify::touch-mode', () => this.syncUI());
        this.pm.connectTo(global.backend.get_monitor_manager(), 'monitors-changed', () => this.syncUI());
        this.syncUI();
    }
    syncUI() {
        let touchMode = Clutter.get_default_backend().get_default_seat().get_touch_mode();
        if (touchMode) {
            this.navigationBar?.show();
        }
        else {
            this.navigationBar?.hide();
        }
        // TODO: uncomment:
        this.virtualTouchpad?.setCanOpen(touchMode /*&&  global.display.get_n_monitors() > 1*/);
    }
    defineFeatures() {
        this.defineFeature('navigation-bar', pm => new NavigationBarFeature(pm), 
        // Note: The below callback is also responsible for 'nulling out' everything when the extension is
        // disabled; it's called with `undefined` as argument then:
        f => this.navigationBar = f, settings.navigationBar.enabled);
        this.defineFeature('osk-key-popups', pm => new OskKeyPopupsFeature(pm), f => this.oskKeyPopups = f, settings.oskKeyPopups.enabled);
        this.defineFeature('floating-screen-rotate-button', pm => new FloatingScreenRotateButtonFeature(pm), f => this.floatingScreenRotateButton = f, settings.screenRotateUtils.floatingScreenRotateButtonEnabled);
        this.defineFeature('notification-gestures', pm => new NotificationGesturesFeature(pm), f => this.notificationGestures = f, settings.notificationGestures.enabled);
        this.defineFeature('donations', pm => new DonationsFeature(pm), f => this.donations = f);
    }
    /**
     * A utility method to define [ExtensionFeature]s that are optionally automatically enabled/disabled
     * depending on the given [setting] and are mapped to a class attribute using [assign].
     *
     * Note that the [assign] callback can (and will upon feature or extension disabling) be
     * called with `undefined' as its value; this is intended behavior and the callback should
     * unset the reference it assigned before in this case.
     *
     * All features are created in a patch and are therefore automatically disabled and set to
     * `undefined` when the extension is disabled.
     *
     * For example usages see [defineFeatures] above.
     */
    defineFeature(featureName, create, assign, setting) {
        let p = this.pm.registerPatch(() => {
            // Create the feature and call `assign` to allow the callee to create references:
            let feature = create(this.pm.fork(featureName));
            assign(feature);
            // Destroy the feature and call assign with `undefined` to remove all references/null out
            // the class attribute when the extension is disabled:
            return () => {
                feature?.destroy();
                feature = undefined;
                assign(undefined);
            };
        }, `enable-feature(${featureName})`);
        if (setting) {
            // Enable the feature initially if setting is set to true:
            if (setting.get())
                p.enable();
            // Connect to setting changes:
            this.pm.connectTo(setting, 'changed', value => {
                if (value) {
                    p.enable();
                    this.syncUI();
                }
                else {
                    p.disable();
                }
            });
        }
        else {
            p.enable();
        }
    }
    disable() {
        // Cancel any pending delays:
        Delay.getAllPendingDelays().forEach(d => d.cancel());
        // Destroy the root PatchManager and with that all its descendents:
        this.pm?.destroy();
        this.pm = undefined;
        TouchUpExtension.instance = undefined;
    }
}

export { TouchUpExtension as default };
