import Gio from 'gi://Gio';
import GLib from 'gi://GLib';

const Methods = Object.freeze({
    verify: 0,
    temporary: 1,
    persistent: 2,
});
function callDbusMethod(method, handler, params = null) {
    if (handler !== null && handler !== undefined) {
        Gio.DBus.session.call('org.gnome.Mutter.DisplayConfig', '/org/gnome/Mutter/DisplayConfig', 'org.gnome.Mutter.DisplayConfig', method, params, null, Gio.DBusCallFlags.NONE, -1, null, handler);
    }
    else {
        Gio.DBus.session.call('org.gnome.Mutter.DisplayConfig', '/org/gnome/Mutter/DisplayConfig', 'org.gnome.Mutter.DisplayConfig', method, params, null, Gio.DBusCallFlags.NONE, -1, null);
    }
}
function setMonitorTransform(transform, targetMonitor) {
    DisplayConfigState.getCurrent()
        .then((state) => {
        targetMonitor ??= state.builtinMonitor ?? state.monitors[0];
        const logicalMonitor = state.getLogicalMonitorFor(targetMonitor.connector);
        if (logicalMonitor) {
            logicalMonitor.transform = transform;
            callDbusMethod('ApplyMonitorsConfig', null, state.packToApply(Methods.temporary));
        }
    });
}
class LogicalMonitor {
    x;
    y;
    scale;
    transform;
    primary;
    monitors;
    properties;
    constructor(variant) {
        const unpacked = variant.unpack();
        this.x = unpacked[0].unpack();
        this.y = unpacked[1].unpack();
        this.scale = unpacked[2].unpack();
        this.transform = unpacked[3].unpack();
        this.primary = unpacked[4].unpack();
        this.monitors = unpacked[5].deep_unpack();
        this.properties = unpacked[6].unpack();
        for (const key in this.properties) {
            this.properties[key] = this.properties[key].unpack().unpack();
        }
    }
}
class Monitor {
    connector;
    currentModeId = null;
    isUnderscanning = false;
    isBuiltin = false;
    constructor(variant) {
        const unpacked = variant.unpack();
        this.connector = unpacked[0].unpack()[0].unpack();
        const modes = unpacked[1].unpack();
        for (const modeVariant of modes) {
            const mode = modeVariant.unpack();
            const id = mode[0].unpack();
            const modeProps = mode[6].unpack();
            if ('is-current' in modeProps) {
                const isCurrent = modeProps['is-current'].unpack().get_boolean();
                if (isCurrent) {
                    this.currentModeId = id;
                    break;
                }
            }
        }
        const props = unpacked[2].unpack();
        if ('is-underscanning' in props) {
            this.isUnderscanning = props['is-underscanning'].unpack().get_boolean();
        }
        if ('is-builtin' in props) {
            this.isBuiltin = props['is-builtin'].unpack().get_boolean();
        }
    }
}
class DisplayConfigState {
    serial;
    monitors = [];
    logicalMonitors = [];
    properties;
    constructor(result) {
        const unpacked = result.unpack();
        this.serial = unpacked[0].unpack();
        const monitorVariants = unpacked[1].unpack();
        for (const monitorPacked of monitorVariants) {
            this.monitors.push(new Monitor(monitorPacked));
        }
        const logicalMonitorVariants = unpacked[2].unpack();
        for (const logicalMonitorPacked of logicalMonitorVariants) {
            this.logicalMonitors.push(new LogicalMonitor(logicalMonitorPacked));
        }
        this.properties = unpacked[3].unpack();
        for (const key in this.properties) {
            this.properties[key] = this.properties[key].unpack().unpack();
        }
    }
    static async getCurrent() {
        return new Promise((resolve, reject) => {
            callDbusMethod('GetCurrentState', (conn, res) => {
                try {
                    const reply = conn?.call_finish(res);
                    const configState = new DisplayConfigState(reply);
                    resolve(configState);
                }
                catch (err) {
                    reject(err);
                }
            });
        });
    }
    get builtinMonitor() {
        return this.monitors.find((monitor) => monitor.isBuiltin) ?? this.monitors[0];
    }
    getMonitor(connector) {
        return this.monitors.find((monitor) => monitor.connector === connector) || null;
    }
    getLogicalMonitorFor(connector) {
        return this.logicalMonitors.find((logMonitor) => logMonitor.monitors.some((lmMonitor) => connector === lmMonitor[0])) || null;
    }
    packToApply(method) {
        const packing = [this.serial, method, [], {}];
        const logicalMonitors = packing[2];
        const properties = packing[3];
        this.logicalMonitors.forEach((logicalMonitor) => {
            const lmonitorPack = [
                logicalMonitor.x,
                logicalMonitor.y,
                logicalMonitor.scale,
                logicalMonitor.transform,
                logicalMonitor.primary,
                [],
            ];
            const monitors = lmonitorPack[5];
            for (const logMonitor of logicalMonitor.monitors) {
                const connector = logMonitor[0];
                const monitor = this.getMonitor(connector);
                if (monitor) {
                    monitors.push([
                        connector,
                        monitor.currentModeId,
                        {
                            enable_underscanning: new GLib.Variant('b', monitor.isUnderscanning),
                        },
                    ]);
                }
            }
            logicalMonitors.push(lmonitorPack);
        });
        if ('layout-mode' in this.properties) {
            properties['layout-mode'] = new GLib.Variant('b', this.properties['layout-mode']);
        }
        return new GLib.Variant('(uua(iiduba(ssa{sv}))a{sv})', packing);
    }
}

export { DisplayConfigState, LogicalMonitor, Methods, Monitor, callDbusMethod, setMonitorTransform };
